#include <stdio.h>
#include <stdlib.h>

/* un edge e' associato a ogni arco uscente da un dato vertice */
/* e appartiene alla lista adj del dato vertice */
typedef struct edge {
	int vertexlabel;        /* indice del vertice in cui l'arco entra */
	int weight;             /* peso dell'arco */
	struct edge *next;
} edge;

/* i vertici sono memorizzati in un array dinamico di vertex */
typedef struct vertex {
	int d;                  /* stima del costo del cammino minimo da 0 a questo vertice */
	int pred;               /* indice del vertice predecessore nel cammino stimato */
	edge *adj;              /* lista di adiacenza */ 
	int priorityindex;      /* indice dell'elemento corrispondente al vertice nella coda di priorita' */
} vertex;

/* struttura per manipolare il grafo */
typedef struct graph {
	int V;                  /* numero vertici */
	int E;                  /* numero archi */
	vertex *A;              /* array dinamico di vertex */
} graph;

/* struttura per manipolare la coda di priorita' */
typedef struct priorityqueue {
	int size;               /* dimensione dello heap */
	int *vertexlabel;       /* array dinamico contenente lo heap */
} priorityqueue;

/* scambia due elementi nella coda */
void vertexswap(graph *G, int *a, int i, int j)
{
	vertex *v = G->A;
	int k = a[i];

	v[a[i]].priorityindex = j;    /* per aggiustare i collegamenti dall'array di vertex verso la coda */
	v[a[j]].priorityindex =	i;
	a[i] = a[j];
	a[j] = k;
}

/* boolean: restituisce true se il campo d del vertex corrispondente all'i-esimo elemento della coda e' */
/*          minore del campo d del vertex corrispondente al j-esimo elemento della coda. */
char cfr(graph *G, int *a, int i, int j)
{
	vertex *v = G->A;
	
	if(v[a[i]].d == -1)           /* -1 rappresenta +infinito */
		return 0;
	if(v[a[j]].d == -1)
		return 1;
	return v[a[i]].d < v[a[j]].d;	
}

/* ripristina la proprieta' dello heap eventualmente violata dall'elemento di indice i */
void heapify(graph *G, int *a, int n, int i)
{
	int l = 2 * i + 1;
	int r = l + 1;
	int m;

	m = (l < n && cfr(G,a,l,i)) ? l : i;
	if(r < n && cfr(G,a,r,m))
		m = r;
	if(m != i) {
		vertexswap(G,a,i,m);
		heapify(G,a,n,m);
	}
}

/* alloca e costruisce la coda di priorita' e lo heap */
/* stabilisce i collegamenti iniziali da heap a array di vertici */
/* e da array di vertici a heap (campo priorityindex) */
priorityqueue *buildheap(graph *G)
{
	int i;
	int n = G->V;
	priorityqueue *pq = malloc(sizeof(priorityqueue));
	
	if(!pq) {
		fprintf(stderr,"Errore di Allocazione\n");
		exit(-4);
	};
	if(!(pq->vertexlabel = calloc(pq->size = n, sizeof(int)))) {
		fprintf(stderr,"Errore di Allocazione\n");
		exit(-5);
	};
	for(i = 0; i < n; i++) {
		pq->vertexlabel[i] = i;
		G->A[i].priorityindex = i;
	}
	for(i = n/2-1; i >= 0; i--)
		heapify(G,pq->vertexlabel,n,i);
	return pq;
}

/* estrae l'elemento minimo dalla coda pq */
int extractminheap(graph *G, priorityqueue *pq)
{
	int *a = pq->vertexlabel;
	int m = a[0];

	a[0] = a[--(pq->size)];
	heapify(G,a,pq->size,0);
	return m;
} 

/* decrementa il valore d di un elemento e dunque deve risistemare la coda */
void decreasevalue(graph *G, priorityqueue *pq, int i, int k)
{	
	int *a = pq->vertexlabel;
	vertex *v = G->A;

	if(v[a[i]].d != -1 && k > v[a[i]].d)
		return;
	v[a[i]].d = k;
	while(i && cfr(G,a,i,(i-1)/2) ) {
		vertexswap(G,a,i,(i-1)/2);
		i = (i-1)/2;
	}
}

/* stampa contenuto della coda e del campo d di ogni elemento della coda */
void stampaq(graph *G, priorityqueue *pq, char *s)
{
	int i;

	puts(s);
	for(i = 0; i < pq->size; i++)
		printf("{%d(%d)} ",pq->vertexlabel[i],G->A[pq->vertexlabel[i]].d);
	putchar('\n');
}

/* boolean: restituisce true se la coda e' vuota */
char empty(priorityqueue *pq)
{
	return !(pq->size);
}

void destroy(priorityqueue *pq)
{
	free(pq->vertexlabel);
	free(pq);
}

/* inizializza il grafo per eseguire Dijkstra */
void initialize(graph *G, int s)
{
	int i;

	for(i = 0; i < G->V; i++)
		G->A[i].d = G->A[i].pred = -1;
	G->A[s].d = 0;
}

/* algoritmo di Dijkstra: calcola un albero di cammini minimi sul grafo G dal vertice s */
void dijkstra(graph *G, int s)
{
	int u, w;
	edge *e;
	vertex *v;
	priorityqueue *pq;

	initialize(G,s);
	pq = buildheap(G);
	while(!empty(pq))  {		
		u = extractminheap(G,pq);
		if(G->A[u].d == -1)
			continue;		
		for(e = G->A[u].adj; e; e = e->next) {
			/* rilassamento dell'arco e */
			v = G->A + e->vertexlabel;	
			w = G->A[u].d + e->weight;
			if(v->d == -1 || v->d > w) {
				decreasevalue(G,pq,v->priorityindex,w);
				v->pred = u;
			}
		}
	}
	destroy(pq);
}

/* stampa il cammino trovato da s a v */
void stampacammino(graph *G, int s, int v)
{
	if(s == v) {
		printf("[%d]",s);
		return;
	}
	if(G->A[v].pred == -1)
		return;
	stampacammino(G,s,G->A[v].pred);
	printf("[%d]",v);
}

/* stampa tutti i cammini trovati dalla sorgente s */
void stampacammini(graph *G, int s)
{
	int i;

	for(i = 0; i < G->V; i++) {
		if(s != i) {
			if(G->A[i].d == -1)
				printf("Non ci sono cammini da %d a %d",s,i);
			else
				printf("Un cammino minimo da %d a %d, di costo %d e':",s,i,G->A[i].d);
			stampacammino(G,s,i);
			putchar('\n');
		}
	}
}

/* alloca la struttura dati che contiene il grafo */
graph *creategraph(int nv, int ne)
{
	graph *g = malloc(sizeof(graph));
	
	if(!g) {
		fprintf(stderr,"Errore di Allocazione\n");
		exit(-1);
	}
	g->E = ne;
	g->V = nv;
	if(!(g->A = calloc(nv,sizeof(vertex))))  {
		fprintf(stderr,"Errore di Allocazione\n");
		exit(-2);
	}
	return g;
}

/* aggiunge un arco alla lista di adiacenza di un vertice */
edge *edgeinsert(edge *p, int k, int w)
{
	edge *q = malloc(sizeof(edge));

	if(!q) {
		fprintf(stderr,"Errore di Allocazione\n");
		exit(-3);
	}
	q->vertexlabel = k;
	q->weight = w;
	q->next = p;
	return q;
}

/* popola il grafo, leggendone la specifica dal file puntato da fp */
void readgraph(graph *g, FILE *fp)
{
	int i, v1, v2, w;
	
	for(i = 0; i < g->E; i++) {
		fscanf(fp,"%d %d %d",&v1,&v2,&w);
		g->A[v1].adj = edgeinsert(g->A[v1].adj,v2,w);
	}		
}

int main(int ac, char **av)
{
	graph *G;
	int v,e,i;
	FILE *fp;
	
	if(ac != 2) {
		fprintf(stderr,"Uso: %s nomefile\n", av[0]);
		exit(-6);
	}
	if(!(fp = fopen(av[1],"r"))) {
		fprintf(stderr,"Non riesco ad aprire il file %s\n",av[1]);
		exit(-7);
	}

	fscanf(fp,"%d %d",&v,&e);
 	G = creategraph(v,e);
	readgraph(G,fp);
	fclose(fp);

	for(i = 0; i < v; i++) {
		putchar('\n');
		dijkstra(G,i);
		stampacammini(G,i);
	}
	
	return 0;
}
